"""
.. _parsers_flowdata_formats:

List of supported file formats:

 - DryCal log file text output (txt) :mod:`yadg.parsers.flowdata.drycal`
 - DryCal log file tabulated output (csv) :mod:`yadg.parsers.flowdata.drycal`
 - DryCal log file document file (rtf) :mod:`yadg.parsers.flowdata.drycal`
"""
from .main import process, version
