"""
.. _parsers_masstrace_formats:

List of supported file formats:

 - Pfeiffer Quadstar 32-bit scan analog data (sac) :mod:`yadg.parsers.masstrace.quadstarsac`
"""
from .main import process, version
